@@ -29,10 +29,10 @@
 
 ################################################################################
 # FUNCTION:               DESCRIPTION:
+#  fARMA                   Class Representation for "fARMA" Objects
 #  armaSim                 Simulates an AIMA Time Series Process
-#   fARMA                   Class Representation for "fARMA" Objects
 #  armaFit                 Fits Model Parameters for ARMA Time Series Process
-# METHODS:                DESCRIPTION:
+# S3 METHOD:              DESCRIPTION:
 #  predict.fARMA           S3: Predicts from an ARMA Time Series Process 
 #  print.fARMA             S3: Prints a Fitted ARMA Time Series Object
 #  plot.fARMA              S3: Plots Stylized Facts of a Fitted ARMA Object
@@ -42,6 +42,7 @@
 # FUNCTION:               DESCRIPTION:
 #  armaTrueacf             True ARMA Autocorrelation Function
 #  armaRoots               Roots of the ARMA Characteristic Polynomial
+#  armaToeplitz            
 ################################################################################
 # BUILTIN - PACKAGE DESCRIPTION:
 #  Package: fracdiff
@@ -72,9 +73,28 @@
 ################################################################################
 
 
+setClass("fARMA", 
+    representation(
+        call = "call",
+        formula = "formula",
+        method = "character",
+        parameter = "list",
+        data = "list",
+        fit = "list",
+        residuals = "numeric",
+        fitted.values = "numeric",
+        title = "character",
+        description = "character")  
+)
+
+
+# ******************************************************************************
+
+
 armaSim = 
 function(model = list(ar = c(0.5, -0.5), d = 0, ma = 0.1), n = 100,
-innov = NULL, n.start = 100, start.innov = NULL, rand.gen = rnorm, ...) 
+innov = NULL, n.start = 100, start.innov = NULL, rand.gen = rnorm, 
+rseed = NULL, ...) 
 {   # A function implemented by Diethelm Wuertz
 
     # Description:
@@ -83,25 +103,43 @@ innov = NULL, n.start = 100, start.innov = NULL, rand.gen = rnorm, ...)
     # Note:
     #   Splus-Like argument list
     
+    # Example:
+    #	armaSim(model = list(ar = c(0.5, -0.5), d = 0, ma = 0.1))
+    #	armaSim(model = list(ar = c(0.5, -0.5), d = 0.2, ma = 0.1))
+    #	armaSim(model = list(ar = 0, d = 0.2, ma = 0))
+    #	armaSim(model = list(d = 0.2))
+    
     # FUNCTION:
     
+    # Checks:
+    if (!is.list(model)) 
+    	stop("model must be a list")
+    	
     # Simulate:
-    if (!is.list(model)) stop("model must be list")
-    if (is.null(innov)) innov = rand.gen(n, ...)
+    if (is.integer(rseed)) 
+    	set.seed(rseed)
+    if (is.null(innov)) 
+    	innov = rand.gen(n, ...)
     n = length(innov) 
-    if (is.null(start.innov)) start.innov = rand.gen(n, ...) 
+    if (is.null(start.innov)) 
+    	start.innov = rand.gen(n, ...) 
     n.start = length(start.innov)
 
     # AR PART:
     p = length(model$ar)
-    if (p == 1 && model$ar == 0) p = 0
-    if (p) { minroots = min(Mod(polyroot(c(1, -model$ar))))
-        if (minroots <= 1) stop("ar part of model is not stationary") }
+    if (p == 1 && model$ar == 0) 
+    	p = 0
+    if (p) { 
+	    minroots = min(Mod(polyroot(c(1, -model$ar))))
+        if (minroots <= 1) stop("ar part of model is not stationary") 
+    }
     
     # MA PART:
     q = length(model$ma)
-    if (q == 1 && model$ar == 0) q = 0
-    if (n.start < p + q) stop("burn-in must be as long as ar + ma")
+    if (q == 1 && model$ma == 0) 
+    	q = 0
+    if (n.start < p + q) 
+    	stop("burn-in must be as long as ar + ma")
     
     # DIFFERENCING:
     ## if (model$d < 0) stop("d must be positive ") 
@@ -109,11 +147,15 @@ innov = NULL, n.start = 100, start.innov = NULL, rand.gen = rnorm, ...)
     if (dd) { 
         # FRACDIFF if "dd" is a non-integer value:
         d = model$d
-        if (d != round(d) ) { TSMODEL = "FRACDIFF" }
-        else { TSMODEL = "ARIMA" } }
-    else {
+        if (d != round(d) ) { 
+	        TSMODEL = "FRACDIFF" 
+	    } else { 
+		    TSMODEL = "ARIMA" 
+	    } 
+	} else {
         d = 0 
-        TSMODEL = "ARIMA" } 
+        TSMODEL = "ARIMA" 
+    } 
     
     # ARMA:
     if (TSMODEL == "ARIMA") {
@@ -121,50 +163,28 @@ innov = NULL, n.start = 100, start.innov = NULL, rand.gen = rnorm, ...)
         if (length(model$ma)) x = filter(x, c(1, model$ma), sides = 1)
         if (length(model$ar)) x = filter(x, model$ar, method = "recursive")
         x = x[-(1:n.start)]
-        if (d > 0) x = diffinv(x, differences = d) }
+        if (d > 0) x = diffinv(x, differences = d) 
+    }
         
     if (TSMODEL == "FRACDIFF") {
-        if (p == 0) ar = 0
-        if (q == 0) ma = 0
+        if (p == 0) model$ar = 0
+        if (q == 0) model$ma = 0
         mu = 0
         # Use Fortran Routine from R's contributed fracdiff package:
         # This is a BUILTIN function ...
         x = .Fortran("fdsim", as.integer(n), as.integer(p), as.integer(q), 
-            as.double(ar), as.double(ma), as.double(d), as.double(mu), 
-            as.double(rnorm(n + q)), x = double(n + q), 
+            as.double(model$ar), as.double(model$ma), as.double(model$d), 
+            as.double(mu), as.double(rnorm(n + q)), x = double(n + q), 
             as.double(.Machine$double.xmin), as.double(.Machine$double.xmax), 
             as.double(.Machine$double.neg.eps), as.double(.Machine$double.eps), 
-            PACKAGE = "fSeries")$x[1:n] }
+            PACKAGE = "fSeries")$x[1:n] 
+    }
                
     # Return Value:
     as.ts(x)
 }
 
 
-# ******************************************************************************
-
-
-require(methods)
-
-
-# ------------------------------------------------------------------------------
-
-
-setClass("fARMA", 
-    representation(
-        call = "call",
-        formula = "formula",
-        method = "character",
-        parameter = "list",
-        data = "data.frame",
-        fit = "list",
-        residuals = "numeric",
-        fitted.values = "numeric",
-        title = "character",
-        description = "character")  
-)
-
-
 # ------------------------------------------------------------------------------
 
 
@@ -173,7 +193,7 @@ function(
 formula = x ~ arima(2, 0, 1),  
 method = c("CSS-ML", "ML", "CSS", "yw", "burg", "ols", "mle"), 
 include.mean = TRUE, fixed = NULL, fracdiff.M = 100, fracdiff.h = -1, 
-title = "", description = "", ...)
+title = NULL, description = NULL, ...)
 {   # A function implemented by Diethelm Wuertz
 
     # Description:
@@ -202,7 +222,9 @@ title = "", description = "", ...)
     if (m != 3) stop("Formula misspecified")
     
     # Get Series:
-    ts = eval(formula[[2]], + sys.parent())
+    # ts = eval(formula[[2]], + sys.parent())
+    # DW 2005-02-04
+    x = ts = eval(formula[[2]], + sys.parent())
     
     # Allow for univariate 'timeSeries' Objects:
     # Added 2004-09-04 DW
@@ -228,6 +250,11 @@ title = "", description = "", ...)
     if (tsmodel == "fracdiff") valid = TRUE
     if (!valid) stop("Invalid Formula Specification")
     
+    
+    # Reset default method properly for AR models if not spefied!
+    # DW 2005-02-04 
+    if (tsmodel == "ar" & method == "CSS-ML") method = "yw"
+    
     # Internal Function: ar
     if (tsmodel == "ar") {
     .arFit <<- function (x, order, include.mean, fixed = NULL,
@@ -252,22 +279,27 @@ title = "", description = "", ...)
         if (include.mean) {
             coeff = c(fit$coef, fit$x.mean)
             names(coeff) = c(names(fit$coef), "intercept") 
-            fit$coef = coeff} 
+            fit$coef = coeff
+        } 
         if (method == "ols") { 
             fit$se.coef = fit$asy.se.coef
             n = sqrt(length(as.vector(fit$se.coef)))
-            fit$var.coef = matrix(rep(NA, times = n*n), ncol = n) }
-        else { 
+            fit$var.coef = matrix(rep(NA, times = n*n), ncol = n) 
+        } else { 
             fit$var.coef = fit$asy.var.coef
             fit$se.coef = sqrt(diag(fit$asy.var.coef))  
             if (include.mean) {        
                 m = dim(fit$asy.var.coef)[1] + 1
                 var.coef = matrix(rep(NA, times = m*m), m, m)
                 for ( i in 1:(m-1) ) { 
-                    for( j in 1:(m-1) ) {
-                        var.coef[i,j] = fit$var.coef[i,j] } }
+                    for ( j in 1:(m-1) ) {
+                        var.coef[i,j] = fit$var.coef[i,j] 
+                    } 
+                }
                 fit$var.coef = var.coef
-                fit$se.coef = c(fit$se.coef, NA) } }
+                fit$se.coef = c(fit$se.coef, NA) 
+            } 
+        }
         fit$x = x
         # Return Value:
         fit } }
@@ -565,7 +597,7 @@ title = "", description = "", ...)
         p = as.integer(substr(order, 1, pos-1))
         q = as.integer(substr(order, pos+1, nchar(order)))
         order = c(p, q)}
-    
+   
     # Fit:
     filter = 100
     fit = fun(x = ts, order = order, include.mean = include.mean, 
@@ -580,7 +612,11 @@ title = "", description = "", ...)
     fit$tsmodel = tsmodel
     fit$class = "fARMA"
     class(fit) = "list"
-    
+       
+    # Add title and desription:
+    if (is.null(title)) title = "ARIMA Modelling"
+    if (is.null(description)) description = as.character(date())
+       
     # Return Value:
     new("fARMA",     
         call = as.call(match.call()),
@@ -588,7 +624,7 @@ title = "", description = "", ...)
         method = as.character(method),
         parameter = list(include.mean = include.mean, fixed = fixed, 
             fracdiff.M = fracdiff.M, fracdiff.h = fracdiff.h),
-        data = as.data.frame(x),
+        data = list(x = x),
         fit = fit,
         residuals = as.vector(fit$residuals),
         fitted.values = as.vector(fit$fitted.values),
@@ -597,7 +633,7 @@ title = "", description = "", ...)
 }
 
 
-# ******************************************************************************
+# ##############################################################################
 
 
 predict.fARMA = 
@@ -853,17 +889,20 @@ function(x, ...)
     
     # FUNCTION:
     
-    # Call:
-    # object$call, object$coef
+    # Fit:
     object = x@fit
     
+    # Title:
+    cat("\nTitle:\n")
+    cat(x@title, "\n")
+    
     # Call:
     cat("\nCall:\n")
     cat(paste(deparse(object$call), sep = "\n", collapse = "\n"), 
         "\n", sep = "")
       
     # Model: 
-    cat("\nModel:\n", object$tstitle, "\n", sep = "")
+    cat("\nModel:\n", object$title, "\n", sep = "")
     
     # Coefficients:
     cat("\nCoefficient(s):\n")
@@ -871,8 +910,11 @@ function(x, ...)
     print.default(format(object$coef, digits = digits), print.gap = 2, 
         quote = FALSE)
         
-    # Return Value
-    cat("\n")
+    # Description:
+    cat("\nDescription:\n")
+    cat(x@description, "\n\n")
+        
+    # Return Value:
     invisible()
 }
 
@@ -928,9 +970,22 @@ function (object, doplot = TRUE, ...)
     if (object$tsmodel == "fracdiff") {
         doplot = FALSE }
     
-    # Print:
-    print.fARMA(x)
-    
+    # Print Part:
+    # Title:
+    cat("\nTitle:\n")
+    cat(x@title, "\n")
+    # Call:
+    cat("\nCall:\n")
+    cat(paste(deparse(object$call), sep = "\n", collapse = "\n"), 
+        "\n", sep = "")
+    # Model: 
+    cat("\nModel:\n", object$tstitle, "\n", sep = "")
+    # Coefficients:
+    cat("\nCoefficient(s):\n")
+    digits = max(4, getOption("digits") - 4) 
+    print.default(format(object$coef, digits = digits), print.gap = 2, 
+        quote = FALSE)
+     
     # Residuals:
     digits = max(4, getOption("digits") - 4)
     if (length(object$residuals) > 2) {
@@ -984,9 +1039,12 @@ function (object, doplot = TRUE, ...)
     # Doplot:
     if (doplot) plot.fARMA(x, ...)
     
+    # Description:
+    cat("\nDescription:\n")
+    cat(x@description, "\n\n")
+    
     # Return Value:
-    cat("\n")
-    invisible(ans)
+    invisible()
 }
 
 
@@ -1173,6 +1231,444 @@ function(coefficients, n.plot = 400, digits = 4, ...)
 }
 
 
-# -----------------------------------------------------------------------------
+################################################################################
+# FUNCTION:          DESCRIPTION:
+#  armaToeplitz       Toeplitz Matrix
+#  armaFischer        ARMA Fischer Information matrix
+# INTERNAL FUNCTION  DESCRIPTION:
+#  .schurTest          Test for invertibility
+#  .toeplitzARMA       Toeplitz matrix
+#  .iARMA              Information matrix ARMA
+#  .iFARMA             Information matrix FARMA
+#  .psiwtsARMA         MA expansion coefficients
+#  .tacvfARMA          True acvf ARMA
+#  .tccfAR             Cross covariances between ARs
+################################################################################
 
 
+armaToeplitz =
+function(x)
+{   # A function implemented by Diethelm Wuertz
+
+	# Wraps:
+	.toeplitzARMA(x)
+}
+
+
+# ------------------------------------------------------------------------------
+
+
+armaFischer = 
+function(model = list(ar = c(0.5, -0.5), ma = 0.1))
+{	# A function implemented by Diethelm Wuertz
+
+	# Wraps:
+	.iARMA(phi = model$ar, theta = model$ma)
+}
+
+
+
+# ******************************************************************************
+
+
+.schurTest = 
+function(phi)
+{   # A function implemented by Diethelm Wuertz
+
+    # Description:
+    #   Tests for invertibility
+    
+    # Details:
+    #   Tests if all roots of 
+    #   1 - phi[1] B - ... - phi[length(phi)] B^length(phi) = 0 
+    #   are outside the unit circle.
+    
+    # References:
+    #   Pagano M., (1973);
+    #       When is an autoregressive process stationary? 
+    #       Commun. in Statist. 1, pp. 533-544.
+    #   McLeod I., (1974);
+    #       Derivation of the Theoretical Autocovariance Function of
+    #       Autoregressive-Moving Average Time Series
+    #       Appl. Statist. 24, pp. 255--256.
+    
+    # Author:
+    #   Original Version from "iarma" R library: A.I. McLeod, July 1998
+    
+    # FUNCTION:
+    
+    # Case 1 - Return Value:
+    if (length(phi) == 0) { 
+        return(TRUE) 
+    }
+        
+    # Case 2 - Return Value:
+    p = length(phi)
+    phi = c(-1, phi)
+    A = matrix(numeric(p^2), nrow = p, ncol = p)
+    for (j in 1:p) {
+        for (i in 1:p) {
+            if (j > i) {
+                A[i, j] = A[j, i] 
+            } else {
+                k = 1:min(i, j)
+                A[i, j] = sum(phi[1 + i - k] * phi[1 + j - k] - 
+                    phi[1 + p + k - i] * phi[1 + p + k - j]) 
+            } 
+        } 
+    }
+    
+    if (dim(A)[1] == attr(chol(A, pivot = TRUE), "rank")) { 
+        return(TRUE) 
+    }
+    
+        
+    # Case 3 - Return Value:    
+    return(FALSE)
+}
+
+
+# ------------------------------------------------------------------------------
+
+
+.toeplitzARMA =
+function(x)
+{   # A function implemented by Diethelm Wuertz
+
+    # Description:
+    #   Computes the Toeplitz Matrix
+    
+    # Details:
+    #   Given a vector x, the Toeplitz matrix is a square matrix of
+    #   order length(x) and with [i,j] entries given by x[abs(i-j)].
+    #   If x is a vector of autocovariances, the Toeplitz matrix is  
+    #   the corresponding covariance matrix of the observatons.
+    
+    # Author:
+    #   Original Version from "iarma" R library: A.I. McLeod, July 1998
+    
+    # FUNCTION:
+    
+    # Compute:
+    ans = matrix(x[1 + abs(outer(seq(along = x), seq(along = x), 
+        FUN = "-"))], byrow = TRUE, ncol = length(x))
+        
+    # Return Value:
+    ans
+}
+
+
+# ------------------------------------------------------------------------------
+
+
+.iARMA = 
+function(phi = numeric(0), theta = numeric(0))
+{   # A function implemented by Diethelm Wuertz
+
+    # Description:
+    #   Computes the Information Matrix of an ARMA Process
+    
+    # Author:
+    #   Original Version from "iarma" R library: A.I. McLeod, July 1998
+    
+    # FUNCTION:
+    
+    # Check:
+    if (!(.schurTest(phi) & .schurTest(theta))) {
+        cat("Model is non-causal or non-invertible\n")
+        return(NULL) 
+    }
+        
+    # Compute:
+    p = length(phi)
+    q = length(theta)
+    unames = vnames = character(0)
+    if (p > 0) {
+        if (p > 1) {
+            vvmatrix = (.tccfAR(phi, phi)[ - (1:(p - 1))])[ - (p + 1)] }
+        else if (p == 1) {
+            vvmatrix = .tccfAR(phi, phi)[ - (p + 1)] }
+        vvmatrix = .toeplitzARMA(vvmatrix)
+        imatrix = vvmatrix
+        vnames = paste("phi(", 1:p, ")", sep = "") 
+    }
+    if (q > 0) {
+        if (q > 1) {
+            uumatrix = (.tccfAR(theta, theta)[ - (1:(q - 1))])[ - ( q + 1)] 
+        } else if (q == 1) {
+            uumatrix = .tccfAR(theta, theta)[ - (q + 1)] 
+        }
+        uumatrix = .toeplitzARMA(uumatrix)
+        imatrix = uumatrix
+        unames = paste("theta(", 1:q, ")", sep = "") 
+    }
+    if (p > 0 && q > 0) {
+        uvmatrix = matrix(numeric(1), nrow = p, ncol = q)
+        tuv =  -.tccfAR(phi, theta)
+        for (i in 1:p) {
+            for (j in 1:q) {
+                uvmatrix[i, j] = tuv[q + i - j] 
+            } 
+        }
+        imatrix = cbind(rbind(vvmatrix, t(uvmatrix)), rbind(uvmatrix, 
+            uumatrix)) 
+    }
+    inames = c(vnames, unames)
+    dimnames(imatrix) = list(inames, inames)
+    
+    # Return Value:
+    imatrix
+}
+
+
+# ------------------------------------------------------------------------------
+
+
+.iFARMA = 
+function(phi = numeric(0), theta = numeric(0), maxlag = 128)
+{   # A function implemented by Diethelm Wuertz
+
+    # Description:
+    #   Computes the Information Matrix of a Fractional ARMA Process
+    
+    # Author:
+    #   Original Version from "iarma" R library: A.I. McLeod, July 1998
+    
+    # FUNCTION:
+    
+    # Internal Functions:
+    .psiwtsAR = function(phi, maxlag){
+        p = length(phi)
+        x = numeric(maxlag + 1)
+        x = 1
+        for (i in 1:p) { 
+	        x[i + 1] = crossprod(phi[1:i], (rev(x))[1:i]) 
+	    }
+        if (maxlag > p) {
+            for (i in (p + 1):maxlag) { 
+                x[i + 1] = crossprod(phi, (rev(x))[1:p]) 
+            } 
+        }
+        x 
+    }
+    
+    jFARMA = function(theta, maxlag) {
+        psis = psiwtsAR(theta, maxlag = maxlag)
+        q = length(theta)
+        J = numeric(q)
+        for (k in 1:q) { 
+	        J[k] = sum(psis/(k + 0:maxlag)) 
+	    }
+        J 
+    }
+
+    # Check
+    if (!(.schurTest(phi) & .schurTest(theta))) {
+        cat("Model is non-causal or non-invertible\n")
+        return(NULL) 
+    }
+        
+    # Compute:  
+    I22 = (pi^2)/6
+    if ((length(phi) == 0) && (length(theta) == 0)) return(I22)
+    I11 = .iARMA(phi = phi, theta = theta)
+    J11 = numeric(0)
+    if (length(phi) > 0) J11 = -jFARMA(phi, maxlag)
+    J12 = numeric(0)
+    if (length(theta) > 0) J12 = jFARMA(theta, maxlag)
+    J = c(J11, J12)
+    I = rbind(I11, J)
+    J = c(J, I22)
+    I = cbind(I, J)
+    inames = c(dimnames(I11)[[1]], "d")
+    dimnames(I) = list(inames, inames)
+    
+    # Return Value:
+    I
+}
+
+
+# ------------------------------------------------------------------------------
+
+
+.psiwtsARMA = 
+function(phi = numeric(0), theta = numeric(0), maxlag = 20)
+{   # A function implemented by Diethelm Wuertz
+
+    # Description:
+    #   Computes MA expansion coefficients
+    
+    # Author:
+    #   Original Version from "iarma" R library: A.I. McLeod, July 1998
+    
+    # FUNCTION:
+    
+    # Compute:
+    r = max((p = length(phi)), (q = length(theta)))
+    phi2 = theta2 = numeric(r)
+    maxlagp1 = 1 + maxlag
+    if (q > 0) theta2[1:q] = theta
+    if (p > 0) phi2[1:p] = phi
+    x = numeric(maxlagp1)
+    x[1] = 1
+    if (r == 0) return(x[1:maxlagp1])
+    for (i in 1:r) {
+        x[i + 1] = crossprod(phi2[1:i], rev(x[1:i])) - theta2[i] 
+    }
+    if (p == 0) return(x[1:maxlagp1])
+    if (maxlag > r) {
+        for (i in (r + 1):maxlag) {
+            x[i + 1] = crossprod(phi, x[i - (0:(p - 1))]) 
+        } 
+    }
+    ans = x[1:maxlagp1]
+            
+    # Return Value:
+    ans
+}
+
+
+# ------------------------------------------------------------------------------
+
+
+.tacvfARMA = 
+function(phi = numeric(0), theta = numeric(0), maxlag = 20)
+{   # A function implemented by Diethelm Wuertz
+
+    # Description:
+    #   Computes theoretical autocovariance function of an ARMA 
+    #   Process
+    
+    # Reference: 
+    #   McLeod A.I. (1975);
+    #   Derivation of the theoretical autocovariance function of 
+    #   autoregressive-moving average models, 
+    #   Applied Statistics 24, pp. 255-256. 
+    
+    # Author:
+    #   Original Version from "iarma" R library: A.I. McLeod, July 1998
+    
+    # FUNCTION:
+    
+    # Check:
+    if (!(.schurTest(phi) & .schurTest(theta))) {
+        cat("Model is non-causal or non-invertible\n")
+        return(NULL) 
+    }
+        
+    # Compute:
+    p = length(phi)
+    q = length(theta)
+    maxlagp1 = maxlag + 1
+    if (max(p, q) == 0) { 
+	    return(c(1, numeric(maxlagp1))) 
+	}
+    r = max(p, q) + 1
+    b = numeric(r)
+    C = numeric(q + 1)
+    C[1] = 1
+    theta2 = c(-1, theta)
+    phi2 = numeric(3 * r)
+    phi2[r] = -1
+    if (p > 0) { 
+        phi2[r + 1:p] = phi 
+    }
+    if (q > 0) {
+        for (k in 1:q) {
+            C[k + 1] =  - theta[k]
+            if (p > 0) {
+                for (i in 1:min(p, k)) {
+                  C[k + 1] = C[k + 1] + phi[i] * C[k + 1 - i] 
+                } 
+            } 
+        } 
+    }
+    for (k in 0:q) {
+        for (i in k:q) {
+            b[k + 1] = b[k + 1] - theta2[i + 1] * C[i - k + 1] 
+        } 
+    }
+    if (p == 0) {
+        g = c(b, numeric(maxlagp1))[1:maxlagp1]
+        return(g) 
+    } else if (p > 0) {
+        a = matrix(numeric(r^2), ncol = r)
+        for (i in 1:r) {
+            for (j in 1:r) {
+                if (j == 1) {
+                  a[i, j] = phi2[r + i - 1] 
+                } else if (j != 1) {
+                  a[i, j] = phi2[r + i - j] + phi2[r + i + j - 2] 
+                } 
+            } 
+        }
+        g = solve(a, -b)
+        if (length(g) <= maxlag) {
+            g = c(g, numeric(maxlag - r))
+            for (i in (r + 1):maxlagp1) {
+                g[i] = phi %*% g[i - 1:p] 
+            }
+            return(g) 
+        } else if (length(g) >= maxlagp1) {
+            return(g[1:maxlagp1]) 
+        } 
+    }
+}
+
+
+# ------------------------------------------------------------------------------
+
+
+.tccfAR = 
+function(phi, theta)
+{   # A function implemented by Diethelm Wuertz
+
+    # Description:
+    #   Computes the theoretical cross-covariance function of two 
+    #   autoregressions
+    #   z[t] - phi[1] z_[t-1] --- phi[p] z[t-p]     = a[t]
+    #   z[t] - theta[1] z_[t-1] --- theta[q] z[t-q] = a[t]
+    #   where p, q are length(phi), length(theta)
+    
+    # Notes:
+    #   Auxilary function used with iarma
+    
+    # Author:
+    #   Original Version from "iarma" R library: A.I. McLeod, July 1998
+    
+    # FUNCTION:
+    
+    # Compute:
+    p = length(phi)
+    q = length(theta)
+    if (p == 0 || q == 0) {
+	    return(numeric(0))
+    }
+    k = p + q
+    rhs = c(-1, rep(0, k - 1))
+    A = matrix(numeric(k^2), nrow = k, ncol = k)
+    for (i in 1:k) {
+        for (j in 1:k) {
+            imj = i - j
+            ijq = i + j - q - 1
+            if (i > q) {
+                if (i > j && imj <= q) {
+	                A[i, j] = theta[imj]
+            	} else if (i > q && imj == 0) A[i, j] = -1 
+            } else {
+                if (ijq > 0 && ijq <= p) {
+                    A[i, j] = phi[ijq]
+                } else if (ijq == 0) 
+                    A[i, j] = -1 
+            }
+        } 
+    } 
+    ans = solve(A, rhs)
+                
+    # Return Value:
+    ans
+}
+
+
+################################################################################
+
